@@ -21,8 +21,10 @@
 #include <stdarg.h>
 #include <string.h>
 
+#ifndef XC_WIN32
 #include <X11/Intrinsic.h>
 #include <X11/StringDefs.h>
+#endif
 
 /*----------------------------------------------------------------------*/
 /* Local includes							*/
@@ -44,24 +46,38 @@
 /*----------------------------------------------------------------------*/
 
 typedef struct {
-   float angle1;
-   float angle2;
-   short radius;
-   short yaxis;
+   float  angle1;
+   float  angle2;
+   short  radius;
+   short  yaxis;
+   XPoint position;
 } arcinfo;
 
-typedef struct {
-   short number;
-   genericptr *element;
-   short *idx;
-} uselection;
+/* Smaller data structure used when saving paths in the undo stacks.    */
+/* We don't need the rendering data from points[].                      */
 
 typedef struct {
+   u_short	type;
+   int		color;
+   eparamptr	passed;
+   u_short	style;
+   float	width;
+   XPoint	ctrl[4];
+} splineinfo;
+
+typedef struct {
+   int number;
+   pointlist points;
+} pathinfo;
+
+typedef struct editelem {
    genericptr element;		/* element being edited */
    union {
       stringpart *string;	/* original contents of string, for label */
       pointlist points;		/* original polygon */
       arcinfo *arcspecs;	/* original arc values */
+      pathinfo *pathspecs;	/* original parts of a path */
+      XPoint	instpos;	/* original position, for an object instance */ 
    } save;
 } editelement;
 
@@ -70,6 +86,8 @@ typedef struct {
    float scale;			/* old scale value */
 } scaleinfo;
 
+u_char undo_collect = (u_char)0;
+
 /*----------------------------------------------------------------------*/
 /* Externally declared variables					*/
 /*----------------------------------------------------------------------*/
@@ -129,7 +147,7 @@ uselection *remember_selection(objinstptr topinst, short *slist, int number)
 }
 
 /*----------------------------------------------------------------------*/
-/* Create a selection list in areawin->from the saved uselection	*/
+/* Create a selection list in areawin from the saved uselection		*/
 /* record.								*/
 /*----------------------------------------------------------------------*/
 
@@ -141,7 +159,8 @@ short *regen_selection(objinstptr thisinst, uselection *srec)
    Boolean reorder = False;
    short *slist;
 
-   slist = (short *)malloc(srec->number * sizeof(short));
+   if (srec->number > 0)
+      slist = (short *)malloc(srec->number * sizeof(short));
 
    k = 0;
    for (i = 0; i < srec->number; i++) {
@@ -162,7 +181,7 @@ short *regen_selection(objinstptr thisinst, uselection *srec)
 	 k++;
       }
       else
-	 Fprintf(stderr, "Error: element 0x%x in select list but not object\n",
+	 Fprintf(stderr, "Error: element %p in select list but not object\n",
 		egen);
    } 
 
@@ -174,21 +193,59 @@ short *regen_selection(objinstptr thisinst, uselection *srec)
    }
 
    if (k == 0) {
-      free(slist);
+      if (srec->number > 0) free(slist);
       return NULL;
    }
    else
       return slist;
 }
 
+/*----------------------------------------------------------------------*/
+/* Use the selection list in the undo record to reorder parts in the	*/
+/* object's part list.  Invert the selection list and return it.	*/
+/*----------------------------------------------------------------------*/
+
+void reorder_selection(Undoptr thisrecord)
+{
+   short *slist, *newlist, snum, i;
+   genericptr *pgen, *plist, egen;
+   objinstptr thisinst = thisrecord->thisinst;
+   objectptr thisobj = thisinst->thisobject;
+
+   snum = (short)thisrecord->idata;
+   slist = (short *)thisrecord->undodata;
+   plist = (genericptr *)malloc(snum * sizeof(genericptr));
+   newlist = (short *)malloc(snum * sizeof(short));
+
+   i = 0;
+   for (pgen = plist; pgen < plist + snum; pgen++) {
+      egen = *(thisobj->plist + i);
+      *(plist + *(slist + i)) = egen;
+      i++;
+   }
+   i = 0;
+   for (pgen = plist; pgen < plist + snum; pgen++) {
+      *(thisobj->plist + i) = *pgen;
+      *(newlist + *(slist + i)) = i;
+      i++;
+   }
+   free(plist);
+   free(thisrecord->undodata);
+   thisrecord->undodata = (char *)newlist;
+}
+
 /*----------------------------------------------------------------------*/
 /* free_editelement ---							*/
 /*									*/
 /*   Free memory allocated to an undo record edit element structure.	*/
 /*----------------------------------------------------------------------*/
 
-void free_editelement(editelement *erec)
+void free_editelement(Undoptr thisrecord)
 {
+   editelement *erec;
+   pathinfo *ppi;
+
+   erec = (editelement *)thisrecord->undodata;
    switch (erec->element->type) {
       case LABEL:
 	 freelabel(erec->save.string);
@@ -199,6 +256,12 @@ void free_editelement(editelement *erec)
       case ARC:
 	 free(erec->save.arcspecs);
 	 break;
+      case PATH:
+	 for (ppi = erec->save.pathspecs; ppi < erec->save.pathspecs +
+		thisrecord->idata; ppi++)
+	    free(ppi->points);
+	 free(erec->save.pathspecs);
+	 break;
    }
    free(erec);
 }
@@ -233,7 +296,7 @@ stringpart *get_original_string(labelptr thislab)
    thisrecord = xobjs.undostack;
    if (thisrecord == NULL) return NULL;
 
-   for (chkrecord = thisrecord->next; chkrecord != NULL; chkrecord = chkrecord->next) {
+   for (chkrecord = thisrecord; chkrecord != NULL; chkrecord = chkrecord->next) {
       switch (chkrecord->type) {
 	 case XCF_Edit:
 	    erec = (editelement *)(chkrecord->undodata);
@@ -245,6 +308,7 @@ stringpart *get_original_string(labelptr thislab)
 	    return NULL;
       }
    }
+   return NULL;  /* yes, this can be reached, if thisrecord->next == NULL */
 }
 
 /*----------------------------------------------------------------------*/
@@ -259,7 +323,7 @@ int select_previous(Undoptr thisrecord)
    Undoptr chkrecord;
    uselection *srec;
 
-   unselect_all();
+   clearselects_noundo();
    for (chkrecord = thisrecord->next; chkrecord != NULL; chkrecord = chkrecord->next) {
 
       /* Selections may cross page changes, but only if in the same series */
@@ -346,13 +410,15 @@ void undelete_one_element(objinstptr thisinst, genericptr thiselem)
 /*									*/
 /*   "mode" is UNDO_MORE if one or more undo records are expected to 	*/
 /*   follow in a series, or UNDO_DONE if this completes a series, or is	*/
-/*   as single undo event.						*/
+/*   as single undo event. The command-line command "undo series start"	*/
+/*   forces all events to be UNDO_MORE until "undo series end" is	*/
+/*   issued.								*/
 /*----------------------------------------------------------------------*/
 
 void register_for_undo(u_int type, u_char mode, objinstptr thisinst, ...)
 {
    va_list args;
-   int usize, drawmode, nval, oval, *idata, snum, deltax, deltay, dir;
+   int drawmode, nval, oval, *idata, snum, deltax, deltay, dir, i;
    short *slist;
    objectptr delobj;
    objinstptr newinst;
@@ -360,7 +426,7 @@ void register_for_undo(u_int type, u_char mode, objinstptr thisinst, ...)
    uselection *srec;
    editelement *erec;
    scaleinfo *escale;
-   genericptr egen, newgen;
+   genericptr egen;
    XPoint *fpoint;
    double scale;
 
@@ -392,7 +458,7 @@ void register_for_undo(u_int type, u_char mode, objinstptr thisinst, ...)
    else
       newrecord->idx = 1;
 
-   if (mode == UNDO_MORE)
+   if (mode == UNDO_MORE || undo_collect > (u_char)0)
       newrecord->idx = -newrecord->idx;
    
    xobjs.undostack = newrecord;
@@ -455,7 +521,7 @@ void register_for_undo(u_int type, u_char mode, objinstptr thisinst, ...)
 
       case XCF_Box: case XCF_Arc: case XCF_Wire: case XCF_Text:
       case XCF_Pin_Label: case XCF_Pin_Global: case XCF_Info_Label:
-      case XCF_Spline: case XCF_Dot: case XCF_Graphic:
+      case XCF_Spline: case XCF_Dot: case XCF_Graphic: case XCF_Join:
 	 /* 1 arg:							*/
 	 /*	egen = element just created (genericptr)		*/
 	 egen = va_arg(args, genericptr);
@@ -486,12 +552,37 @@ void register_for_undo(u_int type, u_char mode, objinstptr thisinst, ...)
 	       erec->save.points =
 			copypoints((pointlist)((splineptr)egen)->ctrl, 4);
 	       break;
+	    case OBJINST:
+	       erec->save.instpos = ((objinstptr)egen)->position;
+	       break;
 	    case ARC:
 	       erec->save.arcspecs = (arcinfo *)malloc(sizeof(arcinfo));
 	       erec->save.arcspecs->angle1 = ((arcptr)egen)->angle1;
 	       erec->save.arcspecs->angle2 = ((arcptr)egen)->angle2;
 	       erec->save.arcspecs->radius = ((arcptr)egen)->radius;
 	       erec->save.arcspecs->yaxis = ((arcptr)egen)->yaxis;
+	       erec->save.arcspecs->position = ((arcptr)egen)->position;
+	       break;
+	    case PATH:
+	       newrecord->idata = ((pathptr)egen)->parts;  /* not needed? */
+	       erec->save.pathspecs = (pathinfo *)malloc(newrecord->idata *
+			sizeof(pathinfo));
+	       for (i = 0; i < newrecord->idata; i++) {
+		  pathinfo *ppi = erec->save.pathspecs + i;
+		  genericptr *pgen = ((pathptr)egen)->plist + i;
+		  switch (ELEMENTTYPE(*pgen)) {
+		     case POLYGON:
+			ppi->number = (TOPOLY(pgen))->number;
+			ppi->points = copypoints((TOPOLY(pgen))->points,
+				ppi->number);
+			break;
+		     case SPLINE:
+			ppi->number = 4;
+			ppi->points = copypoints((pointlist)
+				(TOSPLINE(pgen))->ctrl, 4);
+			break;
+		  }
+	       }
 	       break;
 	 }
 	 newrecord->undodata = (char *)erec;
@@ -499,6 +590,7 @@ void register_for_undo(u_int type, u_char mode, objinstptr thisinst, ...)
 
       case XCF_ChangeStyle:
       case XCF_Justify:
+      case XCF_Color:
 	 /* 2 args:							*/
 	 /*	egen = element that was changed (with new value)	*/
 	 /*	oval = old style value					*/
@@ -508,7 +600,7 @@ void register_for_undo(u_int type, u_char mode, objinstptr thisinst, ...)
 	 newrecord->idata = oval;
 	 break;
 
-      case XCF_Scale:
+      case XCF_Rescale:
 	 /* 2 args:							*/
 	 /*	egen = element that was changed (with new value)	*/
 	 /*	ofloat = old float value				*/
@@ -556,6 +648,15 @@ void register_for_undo(u_int type, u_char mode, objinstptr thisinst, ...)
 	 /* Fprintf(stdout, "Undo: registered move of delta (%d, %d)\n",
 			deltax, deltay); */
 	 break;
+
+       case XCF_Reorder:
+	 /* 2 args:							*/
+	 /*	slist = "before" order of elements			*/
+	 /*	snum = number of elements in list (= # parts)		*/
+	 slist = va_arg(args, short *);
+	 snum = va_arg(args, int);
+	 newrecord->undodata = (char *)slist;
+	 newrecord->idata = snum;
    }
 
    va_end(args);
@@ -568,15 +669,16 @@ void register_for_undo(u_int type, u_char mode, objinstptr thisinst, ...)
 
 short undo_one_action()
 {
-   Undoptr thisrecord, chkrecord;
+  Undoptr thisrecord; /* , chkrecord; (jdk) */
    objectptr thisobj;
    objinstptr thisinst;
    uselection *srec;
    editelement *erec;
    scaleinfo *escale;
    short *slist;
-   XPoint *delta;
-   int i, snum, npage, opage;
+   XPoint *delta, position;
+   int i, j, snum;
+   int savemode;
    float fnum;
    genericptr egen;
    labelptr thislabel;
@@ -585,6 +687,7 @@ short undo_one_action()
    arcptr thisarc;
    splineptr thisspline;
    graphicptr thisgraphic;
+   Boolean need_redraw;
    XCWindowData *savewindow = areawin;
 
    /* Undo the recorded action and shift the undo record pointer.	*/
@@ -592,7 +695,7 @@ short undo_one_action()
    thisrecord = xobjs.undostack;
    if (thisrecord == NULL) {
       Fprintf(stderr, "Nothing to undo!\n");
-      return;
+      return 0;
    }
 
    xobjs.undostack = thisrecord->next;
@@ -608,6 +711,7 @@ short undo_one_action()
    /* Setting eventmode to UNDO_MODE prevents register_for_undo() from 	*/
    /* being called again while executing the event.			*/
 
+   savemode = eventmode;
    eventmode = UNDO_MODE;
 
    /* type-dependent part */
@@ -642,8 +746,10 @@ short undo_one_action()
 	    thisrecord->undodata = NULL;
 	    break;
 	 }
-	 else
+	 else {
+	    /* XXX egen is uninitialized! */
 	    delete_one_element(thisrecord->thisinst, egen);
+	 }
 
 	 /* Put back all the parts */
 	 areawin->selects = thisobj->parts;
@@ -681,13 +787,17 @@ short undo_one_action()
 	 /* If there was a previous selection in the undo list,	*/
 	 /* revert to it.					*/
 	 
+	 need_redraw = (areawin->selects > 0) ? True : False;
 	 select_previous(thisrecord);
-	 draw_all_selected();
+	 if (need_redraw)
+	    drawarea(areawin->area, NULL, NULL);
+	 else
+	    draw_all_selected();
 	 break;
 
       case XCF_Box: case XCF_Arc: case XCF_Wire: case XCF_Text:
 		case XCF_Pin_Label: case XCF_Pin_Global: case XCF_Info_Label:
-		case XCF_Spline: case XCF_Dot: case XCF_Graphic:
+		case XCF_Spline: case XCF_Dot: case XCF_Graphic: case XCF_Join:
 	 egen = (genericptr)thisrecord->undodata;
 	 i = thisrecord->thisinst->thisobject->parts - 1;
 	 if (egen != *(thisrecord->thisinst->thisobject->plist + i)) {
@@ -718,6 +828,31 @@ short undo_one_action()
 	       resolveparams(thisrecord->thisinst);
 	       redrawtext(elab);
 	    } break;
+	    case ARC: {
+	       arcinfo tmpinfo;
+	       arcptr earc = (arcptr)(erec->element);
+	       tmpinfo.angle1 = earc->angle1; 
+	       tmpinfo.angle2 = earc->angle2; 
+	       tmpinfo.radius = earc->radius; 
+	       tmpinfo.yaxis = earc->yaxis; 
+	       tmpinfo.position = earc->position; 
+	       earc->angle1 = erec->save.arcspecs->angle1;
+	       earc->angle2 = erec->save.arcspecs->angle2;
+	       earc->radius = erec->save.arcspecs->radius;
+	       earc->yaxis = erec->save.arcspecs->yaxis;
+	       earc->position = erec->save.arcspecs->position;
+	       *(erec->save.arcspecs) = tmpinfo;
+	       calcarc(earc);
+	       drawarea(areawin->area, NULL, NULL);
+	    } break;
+	    case OBJINST: {
+	       XPoint tmppt;
+	       objinstptr einst = (objinstptr)(erec->element);
+	       tmppt = einst->position;
+	       einst->position = erec->save.instpos;
+	       erec->save.instpos = ((objinstptr)egen)->position;
+	       drawarea(areawin->area, NULL, NULL);
+	    }  break;
 	    case POLYGON: {
 	       pointlist tmppts;
 	       int tmpnum;
@@ -741,19 +876,35 @@ short undo_one_action()
 	       calcspline(espline);
 	       drawarea(areawin->area, NULL, NULL);
 	    } break;
-	    case ARC: {
-	       arcinfo tmpinfo;
-	       arcptr earc = (arcptr)(erec->element);
-	       tmpinfo.angle1 = earc->angle1; 
-	       tmpinfo.angle2 = earc->angle2; 
-	       tmpinfo.radius = earc->radius; 
-	       tmpinfo.yaxis = earc->yaxis; 
-	       earc->angle1 = erec->save.arcspecs->angle1;
-	       earc->angle2 = erec->save.arcspecs->angle2;
-	       earc->radius = erec->save.arcspecs->radius;
-	       earc->yaxis = erec->save.arcspecs->yaxis;
-	       *(erec->save.arcspecs) = tmpinfo;
-	       calcarc(earc);
+	    case PATH: {
+	       pointlist tmppts;
+	       int tmpnum;
+	       polyptr epoly;
+	       splineptr espline;
+	       pathptr epath = (pathptr)(erec->element);
+	       for (i = 0; i < epath->parts; i++) {
+		  genericptr ggen = *(epath->plist + i);
+		  switch (ELEMENTTYPE(ggen)) {
+		     case POLYGON:
+			epoly = (polyptr)ggen;
+		        tmppts = epoly->points;
+			tmpnum = epoly->number;
+			epoly->points = (erec->save.pathspecs + i)->points;
+			epoly->number = (erec->save.pathspecs + i)->number;
+			(erec->save.pathspecs + i)->points = tmppts;
+			(erec->save.pathspecs + i)->number = tmpnum;
+			break;
+		     case SPLINE:
+			espline = (splineptr)ggen;
+			tmppts = copypoints((pointlist)espline->ctrl, 4);
+			for (j = 0; j < 4; j++)
+			   espline->ctrl[j] = *((erec->save.pathspecs + i)->points + j);
+			free((erec->save.pathspecs + i)->points);
+		        (erec->save.pathspecs + i)->points = tmppts;
+			calcspline(espline);
+			break;
+		  }
+	       }
 	       drawarea(areawin->area, NULL, NULL);
 	    }
 	 }
@@ -769,6 +920,7 @@ short undo_one_action()
 	 break;
 
       case XCF_Copy:
+	 clearselects_noundo();
 	 srec = (uselection *)thisrecord->undodata;
 	 slist = regen_selection(thisrecord->thisinst, srec);
 	 thisobj = delete_element(thisrecord->thisinst, slist,
@@ -779,6 +931,7 @@ short undo_one_action()
 	 /* Revert selection to previously selected */
 	 select_previous(thisrecord);
 	 drawarea(areawin->area, NULL, NULL);
+	 draw_all_selected();
 	 break;
 
       case XCF_ChangeStyle:
@@ -810,7 +963,36 @@ short undo_one_action()
 	 drawarea(areawin->area, NULL, NULL);
 	 break;
 
-      case XCF_Scale:
+      case XCF_Color:
+	 /* Color changes */
+	 egen = (genericptr)thisrecord->undodata;
+	 snum = thisrecord->idata;
+	 switch(egen->type) {
+	    case PATH:
+	       thispath = (pathptr)(thisrecord->undodata);
+	       thisrecord->idata = thispath->color;
+	       thispath->color = snum;
+	       break;
+	    case POLYGON:
+	       thispoly = (polyptr)(thisrecord->undodata);
+	       thisrecord->idata = thispoly->color;
+	       thispoly->color = snum;
+	       break;
+	    case ARC:
+	       thisarc = (arcptr)(thisrecord->undodata);
+	       thisrecord->idata = thisarc->color;
+	       thisarc->color = snum;
+	       break;
+	    case SPLINE:
+	       thisspline = (splineptr)(thisrecord->undodata);
+	       thisrecord->idata = thisspline->color;
+	       thisspline->color = snum;
+	       break;
+	 }
+	 drawarea(areawin->area, NULL, NULL);
+	 break;
+
+      case XCF_Rescale:
 	 escale = (scaleinfo *)thisrecord->undodata;
 	 egen = escale->element;
 	 fnum = escale->scale;
@@ -846,28 +1028,41 @@ short undo_one_action()
 	       thisgraphic->scale = fnum;
 	       thisgraphic->valid = FALSE;
 	       break;
+	    case LABEL:
+	       thislabel = (labelptr)egen;
+	       escale->scale = thislabel->scale;
+	       thislabel->scale = fnum;
+	       break;
 	 }
 	 drawarea(areawin->area, NULL, NULL);
 	 break;
 
       case XCF_Flip_X:
-	 areawin->save = *((XPoint *)thisrecord->undodata);
-	 elementflip();
+	 position = *((XPoint *)thisrecord->undodata);
+	 elementflip(&position);
 	 break;
 
       case XCF_Flip_Y:
-	 areawin->save = *((XPoint *)thisrecord->undodata);
-	 elementvflip();
+	 position = *((XPoint *)thisrecord->undodata);
+	 elementvflip(&position);
 	 break;
 
       case XCF_Rotate:
-	 areawin->save = *((XPoint *)thisrecord->undodata);
-	 elementrotate(-thisrecord->idata);
+	 position = *((XPoint *)thisrecord->undodata);
+	 elementrotate(-thisrecord->idata, &position);
 	 break;
 
       case XCF_Move:
 	 delta = (XPoint *)thisrecord->undodata;
+	 select_connected_pins();
 	 placeselects(-(delta->x), -(delta->y), NULL);
+	 reset_cycles();
+	 drawarea(areawin->area, NULL, NULL);
+	 draw_all_selected();
+	 break;
+
+       case XCF_Reorder:
+	 reorder_selection(thisrecord);
 	 drawarea(areawin->area, NULL, NULL);
 	 break;
 
@@ -876,7 +1071,16 @@ short undo_one_action()
 	 break;
    }
 
-   eventmode = NORMAL_MODE;	/* Set on a per-event-type basis? */
+   /* Does this need to be set on a per-event-type basis? */
+   switch (savemode) {
+      case CATALOG_MODE:
+      case CATTEXT_MODE:
+	 eventmode = CATALOG_MODE;
+	 break;
+      default:
+	 eventmode = NORMAL_MODE;
+	 break;
+   }
 
    /* Diagnostic, to check if all multiple-event undo series are resolved */
    if (thisrecord->idx < 0) {
@@ -887,6 +1091,21 @@ short undo_one_action()
    return thisrecord->idx;
 }
 
+/*----------------------------------------------------------------------*/
+/* undo_finish_series ---						*/
+/*	Complete a possibly incomplete undo series by forcing the	*/
+/*	topmost entry to have a positive index.				*/
+/*	Note that for "undo series start|end" to work, undo_collect	*/
+/*	must be set to 0 prior to calling undo_finish_series().		*/
+/*----------------------------------------------------------------------*/
+
+void undo_finish_series()
+{
+   if (undo_collect == (u_char)0)
+      if (xobjs.undostack && xobjs.undostack->idx < 0)
+         xobjs.undostack->idx = -xobjs.undostack->idx;
+}
+
 /*----------------------------------------------------------------------*/
 /* undo_action ---							*/
 /*	Play undo record back to the completion of a series.		*/
@@ -910,11 +1129,12 @@ short redo_one_action()
    objectptr thisobj;
    genericptr egen;
    short *slist;
-   XPoint *delta;
+   XPoint *delta, position;
    uselection *srec;
    editelement *erec;
    scaleinfo *escale;
-   int i, tpage, snum;
+   int i, j, snum;
+   int savemode;
    float fnum;
    labelptr thislabel;
    arcptr thisarc;
@@ -930,7 +1150,7 @@ short redo_one_action()
    thisrecord = xobjs.redostack;
    if (thisrecord == NULL) {
       Fprintf(stderr, "Nothing to redo!\n");
-      return;
+      return 0;
    }
    xobjs.undostack = thisrecord;
    xobjs.redostack = thisrecord->last;
@@ -942,6 +1162,7 @@ short redo_one_action()
       return 0;
    }
 
+   savemode = eventmode;
    eventmode = UNDO_MODE;
 
    /* type-dependent part */
@@ -1020,7 +1241,7 @@ short redo_one_action()
 
       case XCF_Box: case XCF_Arc: case XCF_Wire: case XCF_Text:
       case XCF_Pin_Label: case XCF_Pin_Global: case XCF_Info_Label:
-      case XCF_Spline: case XCF_Dot: case XCF_Graphic:
+      case XCF_Spline: case XCF_Dot: case XCF_Graphic: case XCF_Join:
 	 egen = (genericptr)thisrecord->undodata;
 	 undelete_one_element(thisrecord->thisinst, egen);
 	 drawarea(areawin->area, NULL, NULL);
@@ -1044,6 +1265,31 @@ short redo_one_action()
 	       resolveparams(thisrecord->thisinst);
 	       redrawtext(elab);
 	    } break;
+	    case OBJINST: {
+	       XPoint tmppt;
+	       objinstptr einst = (objinstptr)(erec->element);
+	       tmppt = einst->position;
+	       einst->position = erec->save.instpos;
+	       erec->save.instpos = ((objinstptr)egen)->position;
+	       drawarea(areawin->area, NULL, NULL);
+	    }  break;
+	    case ARC: {
+	       arcinfo tmpinfo;
+	       arcptr earc = (arcptr)(erec->element);
+	       tmpinfo.angle1 = earc->angle1; 
+	       tmpinfo.angle2 = earc->angle2; 
+	       tmpinfo.radius = earc->radius; 
+	       tmpinfo.yaxis = earc->yaxis; 
+	       tmpinfo.position = earc->position; 
+	       earc->angle1 = erec->save.arcspecs->angle1;
+	       earc->angle2 = erec->save.arcspecs->angle2;
+	       earc->radius = erec->save.arcspecs->radius;
+	       earc->yaxis = erec->save.arcspecs->yaxis;
+	       earc->position = erec->save.arcspecs->position;
+	       *(erec->save.arcspecs) = tmpinfo;
+	       calcarc(earc);
+	       drawarea(areawin->area, NULL, NULL);
+	    } break;
 	    case POLYGON: {
 	       pointlist tmppts;
 	       int tmpnum;
@@ -1067,32 +1313,48 @@ short redo_one_action()
 	       calcspline(espline);
 	       drawarea(areawin->area, NULL, NULL);
 	    } break;
-	    case ARC: {
-	       arcinfo tmpinfo;
-	       arcptr earc = (arcptr)(erec->element);
-	       tmpinfo.angle1 = earc->angle1; 
-	       tmpinfo.angle2 = earc->angle2; 
-	       tmpinfo.radius = earc->radius; 
-	       tmpinfo.yaxis = earc->yaxis; 
-	       earc->angle1 = erec->save.arcspecs->angle1;
-	       earc->angle2 = erec->save.arcspecs->angle2;
-	       earc->radius = erec->save.arcspecs->radius;
-	       earc->yaxis = erec->save.arcspecs->yaxis;
-	       *(erec->save.arcspecs) = tmpinfo;
-	       calcarc(earc);
+	    case PATH: {
+	       pointlist tmppts;
+	       int tmpnum;
+	       polyptr epoly;
+	       splineptr espline;
+	       pathptr epath = (pathptr)(erec->element);
+	       for (i = 0; i < epath->parts; i++) {
+		  genericptr ggen = *(epath->plist + i);
+		  switch (ELEMENTTYPE(ggen)) {
+		     case POLYGON:
+			epoly = (polyptr)ggen;
+		        tmppts = epoly->points;
+			tmpnum = epoly->number;
+			epoly->points = (erec->save.pathspecs + i)->points;
+			epoly->number = (erec->save.pathspecs + i)->number;
+			(erec->save.pathspecs + i)->points = tmppts;
+			(erec->save.pathspecs + i)->number = tmpnum;
+			break;
+		     case SPLINE:
+			espline = (splineptr)ggen;
+			tmppts = copypoints((pointlist)espline->ctrl, 4);
+			for (j = 0; j < 4; j++)
+			   espline->ctrl[j] = *((erec->save.pathspecs + i)->points + j);
+			free((erec->save.pathspecs + i)->points);
+		        (erec->save.pathspecs + i)->points = tmppts;
+			calcspline(espline);
+			break;
+		  }
+	       }
 	       drawarea(areawin->area, NULL, NULL);
 	    }
 	 }
 	 break;
 
       case XCF_Flip_X:
-	 areawin->save = *((XPoint *)thisrecord->undodata);
-	 elementflip();
+	 position = *((XPoint *)thisrecord->undodata);
+	 elementflip(&position);
 	 break;
 
       case XCF_Flip_Y:
-	 areawin->save = *((XPoint *)thisrecord->undodata);
-	 elementvflip();
+	 position = *((XPoint *)thisrecord->undodata);
+	 elementvflip(&position);
 	 break;
 
       case XCF_ChangeStyle:
@@ -1124,7 +1386,36 @@ short redo_one_action()
 	 drawarea(areawin->area, NULL, NULL);
 	 break;
 
-      case XCF_Scale:
+      case XCF_Color:
+	 /* Color changes */
+	 egen = (genericptr)thisrecord->undodata;
+	 snum = thisrecord->idata;
+	 switch(egen->type) {
+	    case PATH:
+	       thispath = (pathptr)(thisrecord->undodata);
+	       thisrecord->idata = thispath->color;
+	       thispath->color = snum;
+	       break;
+	    case POLYGON:
+	       thispoly = (polyptr)(thisrecord->undodata);
+	       thisrecord->idata = thispoly->color;
+	       thispoly->color = snum;
+	       break;
+	    case ARC:
+	       thisarc = (arcptr)(thisrecord->undodata);
+	       thisrecord->idata = thisarc->color;
+	       thisarc->color = snum;
+	       break;
+	    case SPLINE:
+	       thisspline = (splineptr)(thisrecord->undodata);
+	       thisrecord->idata = thisspline->color;
+	       thisspline->color = snum;
+	       break;
+	 }
+	 drawarea(areawin->area, NULL, NULL);
+	 break;
+
+      case XCF_Rescale:
 	 escale = (scaleinfo *)thisrecord->undodata;
 	 egen = escale->element;
 	 fnum = escale->scale;
@@ -1160,18 +1451,30 @@ short redo_one_action()
 	       thisgraphic->scale = fnum;
 	       thisgraphic->valid = FALSE;
 	       break;
+	    case LABEL:
+	       thislabel = (labelptr)egen;
+	       escale->scale = thislabel->scale;
+	       thislabel->scale = fnum;
+	       break;
 	 }
 	 drawarea(areawin->area, NULL, NULL);
 	 break;
 
       case XCF_Rotate:
-	 areawin->save = *((XPoint *)thisrecord->undodata);
-	 elementrotate(thisrecord->idata);
+	 position = *((XPoint *)thisrecord->undodata);
+	 elementrotate(thisrecord->idata, &position);
 	 break;
 
       case XCF_Move:
 	 delta = (XPoint *)thisrecord->undodata;
+	 select_connected_pins();
 	 placeselects(delta->x, delta->y, NULL);
+	 reset_cycles();
+	 drawarea(areawin->area, NULL, NULL);
+	 break;
+
+      case XCF_Reorder:
+	 reorder_selection(thisrecord);
 	 drawarea(areawin->area, NULL, NULL);
 	 break;
 
@@ -1180,7 +1483,17 @@ short redo_one_action()
 	 break;
    }
 
-   eventmode = NORMAL_MODE;	/* Set on a per-event-type basis? */
+   /* Does this need to be set on a per-event-type basis? */
+   switch (savemode) {
+      case CATALOG_MODE:
+      case CATTEXT_MODE:
+	 eventmode = CATALOG_MODE;
+	 break;
+      default:
+	 eventmode = NORMAL_MODE;
+	 break;
+   }
+
    areawin = savewindow;
 
    return thisrecord->idx;
@@ -1279,7 +1592,7 @@ void free_undo_data(Undoptr thisrecord, u_char mode)
 
       case XCF_Box: case XCF_Arc: case XCF_Wire: case XCF_Text:
       case XCF_Pin_Label: case XCF_Pin_Global: case XCF_Info_Label:
-      case XCF_Spline: case XCF_Dot: case XCF_Graphic:
+      case XCF_Spline: case XCF_Dot: case XCF_Graphic: case XCF_Join:
 	 /* if MODE_UNDO, the element is on the page, so don't destroy it! */
 	 if (mode == MODE_REDO)
 	    free(thisrecord->undodata);
@@ -1287,7 +1600,7 @@ void free_undo_data(Undoptr thisrecord, u_char mode)
 
       case XCF_Edit:
 	 erec = (editelement *)thisrecord->undodata;
-	 free_editelement(erec);
+	 free_editelement(thisrecord);
 	 break;
 
       case XCF_Copy:
@@ -1305,6 +1618,7 @@ void free_undo_data(Undoptr thisrecord, u_char mode)
       case XCF_Push:
       case XCF_ChangeStyle:
       case XCF_Justify:
+      case XCF_Color:
 	 /* Do nothing --- undodata points to a valid element */
 	 break;
 
@@ -1329,7 +1643,7 @@ void free_undo_data(Undoptr thisrecord, u_char mode)
 
 void free_undo_record(Undoptr thisrecord)
 {
-   Undoptr nextrecord, lastrecord;
+  /* Undoptr nextrecord, lastrecord; (jdk) */
 
    /* Reset the master list pointers */
 
@@ -1357,7 +1671,7 @@ void free_undo_record(Undoptr thisrecord)
 
 void free_redo_record(Undoptr thisrecord)
 {
-   Undoptr nextrecord, lastrecord;
+  /* Undoptr nextrecord, lastrecord; (jdk) */
 
    /* Reset the master list pointers */
 
